<!-- $theme: gaia -->

# Logger sentry

---
# Trouble

- the error information from Logger is **dispersed** among every Elixir node
- it's **difficult** to check all error conveniently
- we need **aggregate** all error information through sentry dashboard

---
# Solutions

- manually
- hack Logger

---
## manually

```elixir
error_msg = "some error information from tubitv elixir project"
Logger.error(error_msg)
Sentry.capture_exception(error_msg, [level: :error, stacktrace: stacktrace])
```

advantage

- straightforward
- ...

disadvantage

- ugly
- not graceful

---
## hack Logger

- hack Logger.error
- console backend
- event model
- init event handler
- send msg to event server
- create one new backend
- the code for sentry backend

---
### hack Logger.error

Logger.error will output the information to Elixir console, and it's api definition is:

![](./resources/6BDACB0F2DF45DA01A7FC4595E7D22FA.jpg)

add callback to error function? NO, can't !

---
### console backend

output information to Elixir console, just because console backend.

[code base for console backend.](https://github.com/elixir-lang/elixir/blob/v1.5.2/lib/logger/lib/logger/backends/console.ex)

key point:

- init
- log event

---

init

```elixir
  defp init(config, state) do
    level = Keyword.get(config, :level)
    device = Keyword.get(config, :device, :user) # key of key
    format = Logger.Formatter.compile Keyword.get(config, :format)
    colors = configure_colors(config)
    metadata = Keyword.get(config, :metadata, []) |> configure_metadata()
    max_buffer = Keyword.get(config, :max_buffer, 32)

    %{state | format: format, metadata: metadata,
              level: level, colors: colors, device: device, max_buffer: max_buffer}
  end
```

log event

```elixir
send(device, {:io_request, self(), ref, {:put_chars, :unicode, output}})
```

---
But who initialized the console backend ? How come the event ?

---
### event model

![IMAGE](./resources/224F9702CA0469434F942A5A1E91E703.jpg)

---
There is Logger event server in Elixir system.

![logger model.png](resources/4F5F2F42434F5B44A20E5D55021212E9.png)

---
### init event handler

```elixir
  def init({mod, handler, args}) do
    case :gen_event.delete_handler(mod, handler, :ok) do
      {:error, :module_not_found} ->
        res = :gen_event.add_sup_handler(mod, handler, args)
        do_init(res, mod, handler)
      _ ->
        init({mod, handler, args})
    end
  end
```

code base: [init event handler through watcher](https://github.com/elixir-lang/elixir/blob/v1.5.2/lib/logger/lib/logger/watcher.ex#L20-L28)

It used gen event model, `gen_event` behaviour will call `Logger.Backends.Console.init/1` when execute `add_sup_handler` function.

---
### send msg to event server

The `Logger.error/1,2` is Macro, it will call

```elixir

  defp notify(:sync, msg),  do: :gen_event.sync_notify(Logger, msg)
  defp notify(:async, msg), do: :gen_event.notify(Logger, msg)
```

finally, and the code is [here](https://github.com/elixir-lang/elixir/blob/v1.5.2/lib/logger/lib/logger.ex#L726-L727).

---
### create one new backend

So, maybe we could create one new backend for our business requirements. And the Elixir allow this:

[the document for backends in Elixir](https://github.com/elixir-lang/elixir/blob/v1.5.2/lib/logger/lib/logger.ex#L184).

> `Logger` supports different backends where log messages are written to. 
> 
> Developers may also implement their own backends, an option that
  is explored in more detail below.


---
### the code for sentry backend

the [code](https://github.com/Tubitv/logger_sentry/blob/0.1.2/lib/sentry.ex) is very simple.

---
# The relationship for Logger and error_logger

Logger is event server in Elixir, it will send event message to Logger event server when we use Logger.`level`/1,2 .

`error_logger` is event server in Erlang, we can ignore it when we write Elixir code. But some error information will send to `error_logger` event server.

- gen process crash info inside Erlang system
- ...

---
## handler for error_logger in Logger

in Logger, there is [one handler](https://github.com/elixir-lang/elixir/blob/v1.5.2/lib/logger/lib/logger/error_handler.ex) for `error_logger`.

---
## handler for error_logger in sentry

in Sentry, there also has [one handler](https://github.com/getsentry/sentry-elixir/blob/6.0.3/lib/sentry/logger.ex) for `error_logger`. 

It will handle error message from `error_logger` event server, and push event to datadog dashboard in more friendly (to sentry dashboard) format.

---
## relationship graph

![](./resources/event_server_relationship.png)

---
# Q&A
